/******************************************************************************
* Module:  Dirtmap.cpp
* Version: 1
* Created: 26.3.2003
* Author:   This file was generated by the XSI shader wizard v1.0
*
* This file contains the definition for Dirtmap.
* Description: Procedurally generated dirt texture map.
*
******************************************************************************/

#include <math.h>
#include <shader.h>

#include "Dirtmap.h"

DLLEXPORT int Dirtmap_version(void)
{
	return 1;
}

/*
 *	Shader initialization function
 */
DLLEXPORT void Dirtmap_init(
	miState * state,
	Dirtmap_params * in_pParams,
	miBoolean *inst_req)
{
	if(!in_pParams) {
		// Set this to miTRUE if you want instance specific initialization.
		*inst_req = miFALSE;
		/* add global init code here */
	} else {
		/* add instance-specific init code here */
	}
}

DLLEXPORT miBoolean Dirtmap
(
	miColor * out_pResult,
	miState * state,
	Dirtmap_params * in_pParams
)
{
	/* put your code here */

	miColor		baseColor;
	miColor		dirtColor;
	miBoolean	probeGeom;
	miBoolean	multEnviro;
	miBoolean	illumInvis;
	miBoolean	bentMode;
	miScalar	nearClip;
	miGeoScalar	nearClipS;
	miScalar	farClip;
	miGeoScalar	farClipS;
	miScalar	raySpread;

	miScalar	dirtGain;

	miVector	rayDir;
	miScalar	rayBlend;

	// A few helper variables needed:
	miScalar	hitDist;		// distance to the first hit
	miColor		hitSum;			// averaged sum of all hits
	miColor		curHit;			// a little helper for color computations
	miColor		curEnvironment; // contains the currently sampled environment
	double		sample[2];		// 2d sample, will be filled by mi_sample
	miInteger	sampleCnt;		// sample counter, needed for mi_sample
	miVector	sampleDir, n;	// unit direction vector ray will be shot at, and the current normal n

	// baseColor can be evaluated here, we'll need it in any case
	baseColor	= *mi_eval_color(&(in_pParams->m_Base));

	// eval those two first - if they are 0, no work has to be done
	miUint		rayCount = (miUint)(*mi_eval_integer(&(in_pParams->m_Samples)));

	// if this is zero, we get to do nothing
	if (rayCount <= 0)
	{
		*out_pResult = baseColor;
		return miTRUE;
	}

	// hoi! if this is true (default), and we got hit by a secondary illum
	// ray, then we can bug out early
	illumInvis  = *mi_eval_boolean(&(in_pParams->m_IllumInvis));

	// it is, it iiiis! let's get outta here...
	if (illumInvis == miTRUE &&
		mihRAY_SECILLUM(state->type) == miTRUE)
	{
		*out_pResult = baseColor;
		return miTRUE;
	}

	// seems like all the early checking didn't save us from doing some
	// real work - eval all (most) other inputs
	dirtColor	= *mi_eval_color(&(in_pParams->m_Dirt));
	probeGeom	= *mi_eval_boolean(&(in_pParams->m_Probe));
	multEnviro  = *mi_eval_boolean(&(in_pParams->m_Environ));
	bentMode	= *mi_eval_boolean(&(in_pParams->m_Bent));
	nearClip	= *mi_eval_scalar(&(in_pParams->m_Near));
	nearClipS	= nearClip;
	farClip		= *mi_eval_scalar(&(in_pParams->m_Far));
	farClipS	= farClip;
	raySpread	= *mi_eval_scalar(&(in_pParams->m_Spread));

	dirtGain	= *mi_eval_scalar(&(in_pParams->m_Gain));

	rayBlend	= *mi_eval_scalar(&(in_pParams->m_DirBlend));

	// we haven't cast any samples, yet
	sampleCnt = 0;

	// so set the hitSum to black
	mih_color_set(&hitSum, 0.0f, 0.0f, 0.0f, 0.0f);

	// set the ray falloff, so we can limit the amount of touched geometry:
	// THIS IS MENTAL RAY 3.2+ ONLY!!!
	mi_ray_falloff(state, &nearClipS, &farClipS);

	// we can precompute that, it never changes in the loop
	n = state->normal;

	// if blend = 0, we always use the normal, so skip this:
	if (rayBlend > 0.0f)
	{
		rayDir = *mi_eval_vector(&(in_pParams->m_Dir));
		// multiply the normal by 1 - blend
		mi_vector_mul(&n, 1.0f - rayBlend);

		// and multiply it with blend
		mi_vector_mul(&rayDir, rayBlend);

		// then add the two together
		mi_vector_add(&n, &rayDir, &n);

		// and normalize again, so we get a new unit direction vector
		mi_vector_normalize(&n);
	}

	// now, pre-multiply the direction vector with spread (this also
	// never changes in the loop)
	mi_vector_mul(&n, 1.0f - (raySpread / 180.0f));		

	// check if we should output bent normals
	if (bentMode)
	{
		while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
		{
			sampleDir = mih_sphere_sample(sample);

			mi_vector_mul(&sampleDir, raySpread / 180.0f);

			mi_vector_add(&sampleDir, &n, &sampleDir);

			// sampleDir now holds our sampling direction, so lets rock

			// while checking for probeGeom here might seem really stupid
			// (if it is true, we return full base color - always!), we
			// never know what the customer wants to do with our nice little
			// shader.
			if (probeGeom && mi_trace_probe(state, &sampleDir, &(state->point)))
			{
				hitDist = (miScalar)(state->child->dist);

				if (hitDist <= nearClip)
				{
					// we hit an object closer than the minimum distance
					// this means we completely ignore the current sample direction for our bent normal
				}
				else if (hitDist >= farClip)
				{
					// we hit an object farther away than the maximum distance
					// this means the sample direction gets fully added to our bent normal
					hitSum.r += sampleDir.x;
					hitSum.g += sampleDir.y;
					hitSum.b += sampleDir.z;
					hitSum.a += 1.0;
				}
				else
				{
					// our hit was somewhere between the minimum and maximum distance
					// this means we use a "weakened" version of sampleDir
					miScalar gain = (miScalar)pow(1.0 - ((hitDist - nearClip) / (farClip - nearClip)), 1.0f / dirtGain);

					hitSum.r += sampleDir.x * gain;
					hitSum.g += sampleDir.y * gain;
					hitSum.b += sampleDir.z * gain;
					hitSum.a += gain;
				}
			}
			else
			{
				// we didn't hit anything at all
				// here we also fully add our sample direction to the bent normal
				hitSum.r += sampleDir.x;
				hitSum.g += sampleDir.y;
				hitSum.b += sampleDir.z;
				hitSum.a += 1.0;
			}
		}

		// hitSum now contains a sum of all (partially) unoccluded sample directions
		// we now normalize this direction, so out_pResult will contain a vector in 
		// the range [[-1,-1,-1],[1,1,1]] in the r, g and b components.
		// the alpha component will hold an estimate of how occluded the sampled hemisphere
		// was: 1.0 means that there was no occlusion at all, 0.0 means that not a single
		// ray "escaped"
		miGeoScalar		hitLen = sqrt(hitSum.r * hitSum.r + hitSum.g * hitSum.g + hitSum.b * hitSum.b);

		out_pResult->r = (miScalar)((miGeoScalar)hitSum.r / hitLen);
		out_pResult->g = (miScalar)((miGeoScalar)hitSum.g / hitLen);
		out_pResult->b = (miScalar)((miGeoScalar)hitSum.b / hitLen);
		out_pResult->a = hitSum.a / sampleCnt;
	}
	else
	{
		// I split this up into two separate cases, so it runs a bit faster
		// first branch gets executed if "Probe Environment" is true, second one of it's false
		if (multEnviro)
		{
			// get a 2-d sample from mi's QMC sampler
			while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
			{
				// map the sample to a sphere
				sampleDir = mih_sphere_sample(sample);

				// premultiply this
				mi_vector_mul(&sampleDir, raySpread / 180.0f);

				// add it together
				mi_vector_add(&sampleDir, &n, &sampleDir);

				// sampleDir now holds our sampling direction, so lets rock

				// first, trace an environment ray
				mi_trace_environment(&curEnvironment, state, &sampleDir);

				// then, send a probe into the scene, but ONLY if probeGeom
				// is true!!! thanks to lazy eval, we can pack that into
				// a single if
				if (probeGeom && mi_trace_probe(state, &sampleDir, &(state->point)))
				{
					// get the distance to the first hit
					hitDist = (miScalar)(state->child->dist);

					// are we below the near threshold?
					if (hitDist <= nearClip)
					{
						// yup, apply full dirt
						mih_color_mul(&curHit, &dirtColor, &curEnvironment);
						mih_color_add(&hitSum, &hitSum, &curHit);
					}
					else if (hitDist >= farClip)
					{
						// we are above the far threshold, so apply full base
						mih_color_mul(&curHit, &baseColor, &curEnvironment);
						mih_color_add(&hitSum, &hitSum, &curHit);
					}
					else
					{
						// we are somewhere inbetween, so get a falloff from 0 - 1
						// also apply the gain, by using pow(falloff, 1 - gain)
						mih_color_mix(&curHit, &baseColor, &dirtColor, (miScalar)pow(1.0 - ((hitDist - nearClip) / (farClip - nearClip)), 1.0f / dirtGain));
						mih_color_mul(&curHit, &curHit, &curEnvironment);
						mih_color_add(&hitSum, &hitSum, &curHit);
					}
				}
				else
				{
					// we didn't hit anything, or geometry probing is disabled, 
					// so apply full base
					mih_color_mul(&curHit, &baseColor, &curEnvironment);
					mih_color_add(&hitSum, &hitSum, &curHit);
				}
			}
		}
		else // ELSE (multEnviro)
		{
			// this is the same again as above, but without respecting the environment
			while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
			{
				sampleDir = mih_sphere_sample(sample);

				mi_vector_mul(&sampleDir, raySpread / 180.0f);

				mi_vector_add(&sampleDir, &n, &sampleDir);

				// sampleDir now holds our sampling direction, so lets rock

				// while checking for probeGeom here might seem really stupid
				// (if it is true, we return full base color - always!), we
				// never know what the customer wants to do with our nice little
				// shader.
				if (probeGeom && mi_trace_probe(state, &sampleDir, &(state->point)))
				{
					hitDist = (miScalar)(state->child->dist);

					if (hitDist <= nearClip)
					{
						mih_color_add(&hitSum, &hitSum, &dirtColor);
						curHit = dirtColor;
					}
					else if (hitDist >= farClip)
					{
						mih_color_add(&hitSum, &hitSum, &baseColor);
						curHit = baseColor;
					}
					else
					{
						mih_color_mix(&curHit, &baseColor, &dirtColor, (miScalar)pow(1.0 - ((hitDist - nearClip) / (farClip - nearClip)), 1.0f / dirtGain));
						mih_color_add(&hitSum, &hitSum, &curHit);
					}
				}
				else
				{
					mih_color_add(&hitSum, &hitSum, &baseColor);
				}
			}
		} //ENDIF (multEnviro)

		// now, divide hitSum through the number of samples, so we get an average
		out_pResult->r = hitSum.r / sampleCnt;
		out_pResult->g = hitSum.g / sampleCnt;
		out_pResult->b = hitSum.b / sampleCnt;
		out_pResult->a = hitSum.a / sampleCnt;
	} // ENDIF (bentMode)

	// reset our mi_ray_fallof to the original values
	mi_ray_falloff(state, &nearClipS, &farClipS);

	// return true, and be happy
	return miTRUE;
}
/*
 *	Shader cleanup function
 */
DLLEXPORT void Dirtmap_exit(
	miState * state,
	Dirtmap_params * in_pParams)
{
	if(!in_pParams) {
		/* add global exit code here */
	} else {
		/* add instance-specific exit code here */
	}
}
